import os
import yaml

from github import Github, Repository, PullRequest


class GitContext(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__dict__ = self


def fetch_token(host, options, context):
    if 'token' in options:
        return options['token']

    gh_hosts = os.path.expanduser('~/.config/gh/hosts.yml')
    if os.path.exists(gh_hosts):
        with open(gh_hosts, 'r') as f:
            gh_hosts = yaml.load(f, Loader=yaml.FullLoader)
        if host in gh_hosts:
            return gh_hosts[host]['oauth_token']
    
    return context.environ.get('GITHUB_TOKEN')


# def create_repo():
#     token = fetch_token(host, options, context)
#     g = Github(token)
#     org = g.get_organization(name)


# def clone_new_branch():
#     pass


# def new_pr():
#     pass



from typing import Dict, List


class ScaffoldContext(dict):
    notes: List[str]
    todos: List[str]
    environ: Dict[str, str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.notes = []
        self.todos = []
        self.environ = {}



import os

from ..context import ScaffoldContext

from ..steps import ScaffoldStep
from ..runtime import ScaffoldRuntime
from ..plugin import ScaffoldPluginContext
from ..rendering import render_text

color = {
    'PURPLE': '\033[35m',
    'CYAN':  '\033[36m',
    'BLUE':  '\033[34m',
    'GREEN':  '\033[32m',
    'YELLOW':  '\033[33m',
    'RED':  '\033[31m',
    'BOLD':  '\033[1m',
    'UNDERLINE':  '\033[4m',
    'ITALIC':  '\033[3m',
    'END':  '\033[0m',
}


def init(context: ScaffoldPluginContext):
    context.add_step('shell', ShellStep())


class ShellStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        commands = render_text(step, context)
        term_colors = dict_to_str(color, 'TERM_%s="%s"\n')
        cmd = """
set +x -ae
%s
%s
""" % (term_colors, commands)
        rc = os.system(cmd)
        if rc != 0:
            raise RuntimeError('Failed to execute command')


def dict_to_str(d, fmt='%s=%s\n'):
    s = ''
    for x in d:
        s += fmt % (x, d[x])
    return s



import click
from click.decorators import help_option

from x_scaffold import engine
from x_scaffold.context import ScaffoldContext
from x_scaffold.runtime import ScaffoldConsoleRuntime

@click.command('apply', help='Apply a new scaffold')
@click.argument('package')
@click.option('-n', '--name', default='xscaffold', help='The name of the scaffold to apply')
def apply_cli(package, name):
    engine.run(ScaffoldContext({}), {
        'package': package,
        'name': name
    }, ScaffoldConsoleRuntime())

@click.group(context_settings=dict(help_option_names=['-h', '--help']))
def cli():
    pass

cli.add_command(apply_cli)

if __name__ == '__main__':
    cli()



from pkgutil import iter_modules
from pathlib import Path
from importlib import import_module


def load_plugins():
    """Load all plugins in the plugins directory"""
    path = get_path_to_plugins()
    path = Path(path)
    modules = []
    for _, name, _ in iter_modules([path]):
        if not name.startswith('__'):
            module = import_module(f'.{name}', package=__name__)
            modules.append(module)
    return modules

def get_path_to_plugins():
    """Get the path to the plugins directory"""
    return Path(__file__).parent.parent / 'plugins'



from typing import Dict
from .runtime import ScaffoldRuntime
from .steps import ScaffoldStep


class ScaffoldPlugin:
    def init(runtime: ScaffoldRuntime):
        pass


class ScaffoldPluginContext(dict):
    steps: Dict[str, ScaffoldStep] = {}
    
    def add_step(self, step_name: str, step: ScaffoldStep):
        self.steps[step_name] = step



from x_scaffold.steps import ScaffoldStep
from github import Github, Repository, PullRequest

import os
import yaml

from ..plugin import ScaffoldPluginContext
from ..context import ScaffoldContext
from ..runtime import ScaffoldRuntime
from ..rendering import render_options


def init(context: ScaffoldPluginContext):
    context.add_step('github_repository', GithubRepoStep())
    context.add_step('github_organization', GithubOrgStep())


def invokeGH(obj, attribute: str, *args, **kwargs):
    return getattr(obj, attribute)(*args, **kwargs)


def fetch_token(options, context):
    if 'token' in options:
        return options['token']

    host = options.get('host', 'github.com')
    gh_hosts = os.path.expanduser('~/.config/gh/hosts.yml')
    if os.path.exists(gh_hosts):
        with open(gh_hosts, 'r') as f:
            gh_hosts = yaml.load(f, Loader=yaml.FullLoader)
        if host in gh_hosts:
            return gh_hosts[host]['oauth_token']
    
    return context.environ.get('GITHUB_TOKEN')


class GithubOrgStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)
        name = options['name']
        token = fetch_token(options, context)
        g = Github(token)
        org = g.get_organization(name)
        gh_steps = options.get('steps', [])
        gh_step: dict
        for gh_step in gh_steps:
            for k, v in gh_step.items():
                invokeGH(org, k, **v)


class GithubRepoStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)
        name = options['name']
        token = fetch_token(options, context)
        g = Github(token)
        repo = g.get_repo(name)
        gh_steps = options.get('steps', [])
        for gh_step in gh_steps:
            for k, v in gh_step.items():
                if 'set_topics' in gh_step:
                    self.set_topics(repo, gh_step['set_topics'], context)
                elif 'add_topics' in gh_step:
                    self.add_topics(repo, gh_step['add_topics'], context)
                elif 'set' in gh_step:
                    repo.edit(**gh_step['set'])
                elif 'branch' in gh_step:
                    self.run_branch(repo, gh_step['branch'], context)
                else:
                    invokeGH(repo, k, **v)
    def run_branch(self, repo: Repository, step, context):
        name = step['name']
        branch = repo.get_branch(name)
        gh_steps = step.get('steps', [])
        for gh_step in gh_steps:
            for k, v in gh_step.items():
                invokeGH(branch, k, **v)
    
    def set_topics(self, repo: Repository, step, context):
        new_topics = step
        repo.replace_topics(new_topics)

    def add_topics(self, repo: Repository, step, context):
        topics = repo.get_topics()
        new_topics = step
        for topic in new_topics:
            if topic not in topics:
                topics.append(topic)
        repo.replace_topics(topics)
    
    def create_pr(self, repo: Repository, step, context):
        opts = render_options(step, context)
        pr: PullRequest.PullRequest = repo.create_pull(**opts)
        context['pr'] = {
            'number': pr.number,
            'url': pr.html_url
        }
    
    def create(self, name: str, github: Github, step, context):
        opts = render_options(step, context)
        parts = name.split('/')
        organization = parts[0]
        name = parts[1]

        github.get_organization(organization).create_repo(name=name, **opts)



from x_scaffold.context import ScaffoldContext
from x_scaffold.runtime import ScaffoldRuntime
from ..steps import ScaffoldStep
from ..plugin import ScaffoldPluginContext
from ..rendering import render_text


def init(context: ScaffoldPluginContext):
    context.add_step("set_context", SetStep())
    context.add_step("add_note", AddNoteStep())
    context.add_step("add_todo", AddTodoStep())


class SetStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        context_names = step
        for context_name in context_names:
            context[context_name] = render_text(
                context_names[context_name], context)


class AddNoteStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: str, runtime: ScaffoldRuntime):
        message = render_text(step, context)
        context.notes.append(message)


class AddTodoStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: str, runtime: ScaffoldRuntime):
        message = render_text(step, context)
        context.todos.append(message)



import json
import os
import yaml

from jinja2 import Environment
from jinja2 import FileSystemLoader

from x_scaffold.context import ScaffoldContext

class RenderUtils(object):  # pylint: disable=R0903
    """Template utilities."""

    @classmethod
    def read_file(cls, path, parse=False):
        """Used to read a file and return its contents."""

        with open(path, 'r') as file_handle:
            if parse:
                parser = get_parser(path)
                return parser.load(file_handle)
            else:
                return file_handle.read()

    @classmethod
    def read_json(cls, path):
        """Used to read a JSON file and return its contents."""

        with open(path, 'r') as file_handle:
            return json.load(file_handle)

    @classmethod
    def read_yaml(cls, path):
        """Used to read a YAML file and return its contents."""

        with open(path, 'r') as file_handle:
            return yaml.load(file_handle, Loader=yaml.FullLoader)


def format_list(value, format='{value}'):
    for idx, x in enumerate(value):
        value[idx] = format.format(value=value[idx], index=idx)
    return value


def yaml_format(value):
    if value is None:
        return 'null'
    return yaml.dump(value, default_flow_style=True)


def json_format(value):
    if value is None:
        return 'null'
    return json.dumps(value)


def get_parser(path):
    ext = os.path.splitext(path)[1]
    if ext == '.yaml' or ext == '.yml':
        return yaml
    elif ext == '.json':
        return json
    else:
        exit('Parser format not supported: %s' % ext)


def render(template_name, context, template_dir):
    """Used to render a Jinja template."""

    env = Environment(loader=FileSystemLoader(template_dir), variable_start_string='${{', variable_end_string='}}')
    env.filters['formatlist'] = format_list
    env.filters['yaml'] = yaml_format
    env.filters['json'] = json_format
    utils = RenderUtils()

    template = env.get_template(template_name)

    return template.render(env=os.environ, utils=utils, **context)


def render_text(text, context: ScaffoldContext):
    """Used to render a Jinja template."""

    env = Environment(variable_start_string='${{', variable_end_string='}}')
    env.filters['formatlist'] = format_list
    env.filters['yaml'] = yaml_format
    env.filters['json'] = json_format
    utils = RenderUtils()

    template = env.from_string(text)

    return template.render(env=context.environ, utils=utils, **context)

def render_value(value, context: ScaffoldContext):
    if isinstance(value, str):
        return render_text(value, context)
    elif isinstance(value, list):
        v_list: list[str] = []
        for x in value:
            v_list.append(render_value(x, context))
        return v_list
    elif isinstance(value, dict):
        opts = value.copy()
        for k, v in opts.items():
            opts[k] = render_value(v, context)
        return opts
    else:
        return value

def render_options(options: dict, context: ScaffoldContext):
    return render_value(options, context)

def render_token_file(path: str, tokens: dict):
    """Used to render a Token File."""

    with open(path, 'r') as file_handle:
        content = file_handle.read()

    return render_tokens(content, tokens)

def render_tokens(content: str, tokens: dict):
    """Used to render a Token File."""

    for token in tokens:
        content = content.replace(token, tokens[token])
    return content


from re import T
from x_scaffold.context import ScaffoldContext
from x_scaffold.runtime import ScaffoldRuntime
from x_scaffold.steps import ScaffoldStep
from ..plugin import ScaffoldPluginContext
from ..engine import execute_scaffold
from ..rendering import render_options


def init(context: ScaffoldPluginContext):
    context.add_step('scaffold', Scaffolder())

class Scaffolder(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        options = render_options(step, context)
        orig_package = context['__package']
        has_private_context = False
        if 'context' in options:
            has_private_context = True
            actual_context = ScaffoldContext(options['context'])
        else:
            actual_context = context
        execute_scaffold(actual_context, options, runtime)
        if has_private_context:
            context.todos.extend(actual_context.todos)
            context.notes.extend(actual_context.notes)
        context['__package'] = orig_package



from getpass import getpass
from typing import Dict, List
import sys
import click

from x_scaffold.context import ScaffoldContext

CLI_COLORS = {
    'PURPLE': '\033[35m',
    'CYAN':  '\033[36m',
    'BLUE':  '\033[34m',
    'GREEN':  '\033[32m',
    'YELLOW':  '\033[33m',
    'RED':  '\033[31m',
    'BOLD':  '\033[1m',
    'UNDERLINE':  '\033[4m',
    'ITALIC':  '\033[3m',
    'END':  '\033[0m',
}

class ScaffoldRuntime:
    def write(self, message: str):
        pass

    def ask(self, prompt):
        pass

class ScaffoldConsoleRuntime(ScaffoldRuntime):
    def log(self, message):
        self.write(message + '\n')

    def ask(self, prompt):
        name = prompt.get('name')
        description = prompt.get('description', name)
        if 'choices' in prompt:
            choices = prompt['choices']
            opts = []
            max_len = 0
            for c in choices:
                if isinstance(c, str):
                    opts.append({
                        't': c,
                        'kw': [c],
                        'v': c
                    })
                else:
                    keywords = ', '.join(c['keywords'])
                    if len(keywords) > max_len:
                        max_len = len(keywords)
                    opts.append({
                        'kw': c['keywords'],
                        'kwt': keywords,
                        't': c['text'],
                        'v': c.get('value', c['text'])
                    })

            self.write(description + ': \n')
            for opt in opts:
                if 'kwt' in opt:
                    opt['kwt'] = opt['kwt'].ljust(max_len)
                    self.write('  - [{kwt}] {t}\n'.format(**opt))
                else:
                    self.write(f'  - {opt["t"]}\n')

            choice = input('Choose: ')
            while True:
                for c in opts:
                    if choice in c['kw']:
                        return c['v']
                self.write('{RED}[invalid choice]{END} Choose: ')
                choice = sys.stdin.readline().strip()

        if prompt.get('secure', False):
            return getpass(prompt=description + ': ')
        else:
            self.write(description + ': ')
            return sys.stdin.readline().strip()
    
    def write(self, message: str):
        sys.stdout.write(message.format(**CLI_COLORS))
        sys.stdout.flush()

    def print_todos(self, context: ScaffoldContext):
        self.write('{BLUE}{BOLD}TODO:{END}\n')
        for todo in context.todos:
            self.write(f'  - {todo}\n')

    def print_notes(self, context: ScaffoldContext):
        self.write('{GREEN}{BOLD}NOTES:{END}\n')
        for note in context.notes:
            self.write(f'  {note}\n')




#!/usr/bin/env python
# coding: utf-8

import argparse
import io
import os
import pathlib

import yaml
import logging
import getpass
import json
import sys
import re
import tempfile
import importlib
from collections import defaultdict
from fnmatch import fnmatch
import git

from .context import ScaffoldContext
from .runtime import ScaffoldRuntime
from .plugins import load_plugins
from .rendering import render_text
from .steps import ScaffoldStep
from .plugin import ScaffoldPluginContext

_log = logging.getLogger(__name__)


def complete(text, state):
    if str(text).startswith('~/'):
        home = os.path.expanduser('~/')
        p = os.path.join(home, text[2:])
    else:
        p = text
        home = None

    items = pathlib.Path(os.getcwd()).glob(p + '*')
    if items is not None and home is not None:
        items = ['~/' + x[len(home):] for x in items]
    return (items + [None])[state]


def set_readline():
    try:
        import readline
        readline.set_completer_delims(' \t\n;')
        readline.parse_and_bind("tab: complete")
        readline.set_completer(complete)
    except:
        pass


class AttributeDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


color = AttributeDict({
    'PURPLE': '\033[35m',
    'CYAN':  '\033[36m',
    'BLUE':  '\033[34m',
    'GREEN':  '\033[32m',
    'YELLOW':  '\033[33m',
    'RED':  '\033[31m',
    'BOLD':  '\033[1m',
    'UNDERLINE':  '\033[4m',
    'ITALIC':  '\033[3m',
    'END':  '\033[0m',
})


def dict_to_str(d, fmt='%s=%s\n'):
    s = ''
    for x in d:
        s += fmt % (x, d[x])
    return s


def str2bool(v):
    if v is None:
        return False
    return v.lower() in ("yes", "true", "t", "1", "y")


known_types = {
    'int': int,
    'bool': str2bool,
    'str': str,
    'float': float
}


def term_color(text, *text_colors):
    return ''.join(text_colors) + text + color.END




# def render_file(path, context):
#     """Used to render a Jinja template."""

#     template_dir, template_name = os.path.split(path)
#     return render(template_name, context, template_dir)


def is_enabled(options):
    if 'enabled' in options:
        return options['enabled']
    if 'disabled' in options:
        return not options['disabled']
    if 'enabledif' in options:
        enabledif = options['enabledif']
        value = enabledif['value']
        if 'equals' in enabledif:
            return value == enabledif['equals']
        elif 'notequals' in enabledif:
            return value != enabledif['notequals']
    return True


def read_input(s):
    return input(s)


# class Prompt:

#     def __init__(self, d):
#         self._dict = d
#         self._value = None

#     def get(self):
#         if self._value is None:
#             self._value = self._get_value(self._dict)
#         return self._value

#     def _get_value(self, prompt):
#         default = prompt.get('default', None)
#         if isinstance(default, str):
#             default = default.format(env=os.environ)

#         if not is_enabled(prompt):
#             return default

#         required = prompt.get('required', False)
#         while True:
#             s = term_color('%s: ' % prompt['description'].format(
#                 default=default, env=os.environ), color.BOLD)
#             # if 'description' in prompt:
#             #     desc = term_color('%s' % prompt['description'], color.ITALIC)
#             #     sys.stdout.write('%s\n' % desc)

#             if 'choices' in prompt:
#                 s = term_color('%s: ' % prompt['description'].format(
#                     default=default), color.BOLD)
#                 sys.stdout.write('%s\n\n' % s)

#                 choices = prompt['choices']
#                 while True:
#                     opts = []
#                     max_len = 0
#                     for c in choices:
#                         keywords = ', '.join(c['keywords'])
#                         if len(keywords) > max_len:
#                             max_len = len(keywords)
#                         opts.append({'kw': keywords, 't': c['text']})

#                     for opt in opts:
#                         s = '%s' % c['text']
#                         opt['kw'] = opt['kw'].ljust(max_len)
#                         sys.stdout.write('[{kw}] {t}\n'.format(**opt))

#                     d = read_input(term_color('\nchoice: ', color.BOLD))

#                     for c in choices:
#                         if d in c['keywords']:
#                             v = c.get('value', d)
#                             if isinstance(v, dict):
#                                 v = defaultdict(
#                                     lambda: '', c.get('default', {}), **v)
#                             return v

#                     sys.stdout.write('\n%s please select a keyword on the left\n\n' %
#                                      term_color('[invalid choice] ', color.RED))
#             else:
#                 if prompt.get('secure', False):
#                     d = getpass.getpass(prompt=s)
#                 else:
#                     d = read_input(s)

#             if d == '' or d is None:
#                 if not required:
#                     return default
#                 else:
#                     sys.stdout.write(term_color('[required] ', color.RED))
#             else:
#                 if 'validate' in prompt:
#                     matches = re.match(prompt['validate'], d)
#                     if matches is None:
#                         sys.stdout.write(term_color(
#                             '[invalid, %s] ' % prompt['validate'], color.RED))
#                         continue
#                 if 'load' in prompt:
#                     if prompt['load'] == 'yaml':
#                         with open(d, 'r') as fhd:
#                             return yaml.load(fhd, Loader=yaml.FullLoader)
#                 return convert(d, prompt.get('type', 'str'))


# class ScaffoldLoader(yaml.Loader):

#     def __init__(self, stream):
#         if stream is not None:
#             if isinstance(stream, io.FileIO):
#                 self._root = os.path.split(stream.name)[0]
#             elif isinstance(stream, dict):
#                 d = stream
#                 stream = d['fhd']
#                 self._root = os.path.splitext(stream.name)[0]
#                 self._context = d['context']

#         super(ScaffoldLoader, self).__init__(stream)

#     def module(self, node):
#         item = self.construct_mapping(node, 9999)

#         if not is_enabled(item):
#             return None

#         fn = load_module(item)
#         if 'args' in item:
#             return fn(**item['args'])
#         else:
#             return fn()

    

#     def prompt(self, node):
#         item = self.construct_mapping(node, 9999)

#         return Prompt(item).get()

#     def prompt2(self, node):
#         item = self.construct_mapping(node, 9999)

#         return Prompt(item)


# ScaffoldLoader.add_constructor('!prompt2', ScaffoldLoader.prompt2)
# ScaffoldLoader.add_constructor('!prompt', ScaffoldLoader.prompt)
# ScaffoldLoader.add_constructor('!module', ScaffoldLoader.module)


def convert(v, type):
    if type in known_types:
        return known_types[type](v)
    return str(v)


def read_parameter(prompt, context, runtime: ScaffoldRuntime):
    default = prompt.get('default', None)
    if isinstance(default, str):
        default = default.format(env=os.environ)

    if not is_enabled(prompt):
        return default

    name = prompt.get('name', 'parameter')
    required = prompt.get('required', False)
    description = prompt.get('description', name)

    if 'if' in prompt:
        enabled = render_text(prompt['if'], context)
        if enabled.lower() != 'true':
            return default
    
    while True:
        # s = term_color('%s: ' % description.format(
        #     default=default, env=os.environ), color.BOLD)
        # if 'description' in prompt:
        #     desc = term_color('%s' % prompt['description'], color.ITALIC)
        #     sys.stdout.write('%s\n' % desc)

        # if 'choices' in prompt:
        #     s = term_color('%s: ' % description.format(
        #         default=default), color.BOLD)
        #     sys.stdout.write('%s\n\n' % s)

        #     choices = prompt['choices']
        #     while True:
        #         opts = []
        #         max_len = 0
        #         for c in choices:
        #             keywords = ', '.join(c['keywords'])
        #             if len(keywords) > max_len:
        #                 max_len = len(keywords)
        #             opts.append({'kw': keywords, 't': c['text']})

        #         for opt in opts:
        #             s = '%s' % c['text']
        #             opt['kw'] = opt['kw'].ljust(max_len)
        #             sys.stdout.write('[{kw}] {t}\n'.format(**opt))

        #         d = None #read_input(term_color('\nchoice: ', color.BOLD))


        #         for c in choices:
        #             if d in c['keywords']:
        #                 v = c.get('value', d)
        #                 if isinstance(v, dict):
        #                     v = defaultdict(
        #                         lambda: '', c.get('default', {}), **v)
        #                 return v

        #         sys.stdout.write('\n%s please select a keyword on the left\n\n' %
        #                             term_color('[invalid choice] ', color.RED))
        # else:
        d = runtime.ask(prompt)

        if d == '' or d is None:
            if not required:
                return default
            else:
                sys.stdout.write(term_color('[required] ', color.RED))
        else:
            if 'validate' in prompt:
                matches = re.match(prompt['validate'], d)
                if matches is None:
                    sys.stdout.write(term_color(
                        '[invalid, %s] ' % prompt['validate'], color.RED))
                    continue
            if 'load' in prompt:
                if prompt['load'] == 'yaml':
                    with open(d, 'r') as fhd:
                        return yaml.load(fhd, Loader=yaml.FullLoader)
            return convert(d, prompt.get('type', 'str'))


def config_cli(args):
    options = {}
    scaffold_file = os.path.expanduser('~/.xscaffold')
    if os.path.exists(scaffold_file):
        with open(scaffold_file, 'r') as fhd:
            options = yaml.load(fhd, Loader=yaml.FullLoader)

    if args.action == 'save':
        options['url'] = args.url

        with open(scaffold_file, 'w') as fhd:
            yaml.dump(options, fhd, default_flow_style=False)
    elif args.action == 'view':
        sys.stdout.write('url: %s' % options.get('url', 'not defined'))


# def log(s, context={}):
#     sys.stdout.write(
#         render_text(s, context)
#     )


# def execute_command(context, pkg_dir, commands):
#     cmds = commands.format(**context)
#     term_colors = dict_to_str(color, 'TERM_%s="%s"\n')
#     cmd = """
# set +x -ae
# %s
# %s
# """ % (term_colors, cmds)
#     rc = os.system(cmd)
#     if rc != 0:
#         raise RuntimeError('Failed to execute command')


# def load_module(m):
#     mod = importlib.import_module('modules.%s' % m['name'])
#     if hasattr(mod, 'init'):
#         getattr(mod, 'init')(sys.modules[__name__])

#     return getattr(mod, m.get('function', 'execute'))


# def execute_modules(context, pkg_dir, modules):
#     for m in modules:
#         execute_fn = load_module(m)
#         execute_fn(context, pkg_dir, m)


def rm_rf(d):
    for path in (os.path.join(d, f) for f in os.listdir(d)):
        if os.path.isdir(path):
            rm_rf(path)
        else:
            os.unlink(path)
    os.rmdir(d)


def process_prompts(d):
    pass
    # for x in d:
    #     if isinstance(d[x], Prompt):
    #         d[x] = read_parameter(d[x].get()
    #     elif isinstance(d[x], dict):
    #         process_prompts(d[x])

def locate_scaffold_file(path, name):
    paths = [
        os.path.join(path, f'{name}.yml'),
        os.path.join(path, f'{name}.yaml'),
        os.path.join(path, f'{name}.json')
    ]
    for p in paths:
        if os.path.exists(p):
            return p
    return None

def process_parameters(parameters, context: ScaffoldContext, runtime: ScaffoldRuntime):
    for parameter in parameters:
        parameter_name = parameter['name']
        if parameter_name in context:
            context[parameter_name] = context[parameter_name]
        else:
            context[parameter_name] = read_parameter(parameter, context, runtime)


def run(context: ScaffoldContext, options, runtime: ScaffoldRuntime):
    execute_scaffold(context, options, runtime)

    runtime.print_todos(context)
    runtime.print_notes(context)

    return context


def execute_scaffold(context: ScaffoldContext, options, runtime: ScaffoldRuntime):
    tempdir = options.get('temp', tempfile.gettempdir())
    package = options['package']

    name = options.get('name', 'xscaffold')

    if os.path.exists(package):
        runtime.log(
            '[info] using local package \'%s\'...' % package + '\n')
        pkg_dir = package
    else:
        pkg_dir = fetch_git(runtime, tempdir, package)

    sys.path.append(pkg_dir)
    scaffold_file = locate_scaffold_file(pkg_dir, name)
    _log.debug('scaffold file: %s', scaffold_file)

    if scaffold_file is not None:
        with open(scaffold_file, 'r') as fhd:
            config = yaml.load(fhd, Loader=yaml.FullLoader)
    else:
        config = {
            'steps': options.get('steps', [{ 'fetch': {} }])
        }

    plugin_context = ScaffoldPluginContext(
        config.get('plugins', {})
    )
    plugins: list = load_plugins()
    for plugin in plugins:
        plugin.init(plugin_context)

    context.update(config.get('context', {}))

    process_parameters(config.get('parameters', []), context, runtime)

    context['__package'] = {
        'path': pkg_dir,
        'options': options
    }

    steps: list = config.get('steps', [])
    step: dict

    invalid_stepnames = ['if']
    for step in steps:
        for step_name in step:
            if step_name in plugin_context.steps:
                if step_name not in invalid_stepnames:
                    if 'if' in step:
                        enabled = render_text(step['if'], context)
                        if enabled.lower() == 'false':
                            continue
                    plugin_step = plugin_context.steps[step_name]
                    plugin_step.run(context, step[step_name], runtime)

    return context


def fetch_git(runtime, tempdir, package):
    package_parts = package.split('@')
    if len(package_parts) == 1:
        package_name = package_parts[0]
        package_version = 'main'
    else:
        package_name = package_parts[0]
        package_version = package_parts[1]
    package_name_parts = package_name.split('/')
    if len(package_name_parts) <= 2:
        package_name_parts = ['github.com'] + package_name_parts
        package_name = '/'.join(package_name_parts)
    pkg_dir = os.path.join(tempdir, f'{package_name}@{package_version}')

    rc = 9999
    if os.path.exists(pkg_dir):
        runtime.log('{YELLOW}[git] updating %s package...{END}\n' % package)
        rc = os.system(
                """(cd {pkg_dir} && git pull >/dev/null 2>&1)""".format(pkg_dir=pkg_dir))
        if rc != 0:
            runtime.log('{RED}[error]{YELLOW} package %s is having issues, repairing...{END}\n' % package)
            rm_rf(pkg_dir)

    if rc != 0:
        runtime.log('[git] pulling %s package...' % package + '\n')
        rc = os.system(f"""
        git clone https://{package_name} {pkg_dir} >/dev/null 2>&1
        """)
    if rc != 0:
        raise Exception(
                'Failed to pull scaffold package %s' % package)

    rc = os.system(f"""(cd {pkg_dir} && git checkout -f {package_version} >/dev/null 2>&1)""")
    if rc != 0:
        raise Exception('Failed to load version %s' % package_version)
    return pkg_dir



from x_scaffold.context import ScaffoldContext
from ..runtime import ScaffoldRuntime
from ..steps import ScaffoldStep
from ..plugin import ScaffoldPluginContext
from ..rendering import render_text


def init(context: ScaffoldPluginContext):
    context.add_step('log', LogStep())


class LogStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        message = render_text(step, context)
        runtime.log(message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['x_scaffold', 'x_scaffold.plugins']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.24,<4.0.0',
 'Jinja2>=3.0.3,<4.0.0',
 'PyGithub>=1.55,<2.0',
 'PyYAML>=6.0,<7.0',
 'click>=7.0.0,<8.0.0']

entry_points = \
{'console_scripts': ['xscaffold = x_scaffold.cli:cli']}

setup_kwargs = {
    'name': 'x-scaffold',
    'version': '2.0.0',
    'description': 'Used to scaffold directories or GitHub repositories.',
    'long_description': None,
    'author': 'Dan Clayton',
    'author_email': 'dan@azwebmaster.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import os
import pathlib

from ..context import ScaffoldContext

from ..rendering import render, render_options, render_token_file, render_tokens
from ..steps import ScaffoldStep
from ..runtime import ScaffoldRuntime
from ..plugin import ScaffoldPluginContext

from fnmatch import fnmatch


def init(context: ScaffoldPluginContext):
    context.add_step('fetch', FetchStep())


def is_match(path, patterns):
    for pattern in patterns:
        if fnmatch(path, pattern):
            return True
    return False


def render_file(path, context):
    """Used to render a Jinja template."""

    template_dir, template_name = os.path.split(path)
    return render(template_name, context, template_dir)


class FetchStep(ScaffoldStep):
    def run(self, context: ScaffoldContext, step: dict, runtime: ScaffoldRuntime):
        opts = render_options(step, context)
        pkg_dir = context['__package']['path']
        full_pkg_dir = os.path.realpath(pkg_dir)
        
        target = opts.get('target', '.')
        target_base = os.path.realpath(context.get('__target', '.'))
        full_target = os.path.realpath(os.path.join(target_base, target))
        source = opts.get('source', '**/*')

        templates = opts.get('templates', [])
        exclude = opts.get('exclude', []) + ['.git', '.git/*']

        paths = pathlib.Path(full_pkg_dir).rglob(source)

        for p_obj in paths:
            p = str(p_obj)
            tfile = p[len(full_pkg_dir)+1:]
            t = os.path.join(full_target, tfile)
            tbase, tname = os.path.split(t)
            if is_match(tfile, exclude):
                continue
            if not os.path.exists(tbase):
                os.makedirs(tbase)

            if p_obj.is_file():
                template = get_template(tfile, templates)
                if template:
                    if 'tokens' in template:
                        content = render_token_file(p, template['tokens'])
                    else:
                        content = render_file(p, context)
                    with open(t, 'w') as fhd:
                        fhd.write(content)
                else:
                    with open(p, 'r') as fhd:
                        with open(t, 'w') as fhd2:
                            fhd2.write(fhd.read())

def get_template(path, templates):
    for template in templates:
        if fnmatch(path, template['path']):
            return template
    return None
